#lang racket

(require racket/fixnum)

(define (get-time) (/ (current-inexact-milliseconds) 1000))

(define ? fxvector-ref)
(define ! fxvector-set!)
 
(define (Merge Source Left1 Right1 Left2 Right2 Target Left)
  (define (copy1)
    (! Target Left (? Source Left1))
    (set! Left (fx+ Left 1))
    (set! Left1 (fx+ Left1 1)))
  (define (copy2)
    (! Target Left (? Source Left2))
    (set! Left (fx+ Left 1))
    (set! Left2 (fx+ Left2 1)))
  (do ()((or (fx> Left1 Right1) (fx> Left2 Right2)))
    (if (fx< (? Source Left1) (? Source Left2)) (copy1) (copy2)))
  (do ()((fx> Left1 Right1)) (copy1))
  (do ()((fx> Left2 Right2)) (copy2)))

(define (Sort Source Left Right Target)
  (cond
    ((fx< Left Right)
     (let ((mid (quotient (fx+ Left Right) 2)))
       (Sort Target Left mid Source)
       (Sort Target (fx+ mid 1) Right Source)
       (Merge Source Left mid (fx+ mid 1) Right Target Left)))))

(define (BinarySearch Value Vector Left Right)    
  (if (fx> Left Right) 
      Left
      (do ((left Left)
           (right (fx+ Right 1))) ((fx>= left right) right)
        (let ((mid (quotient (fx+ left right) 2)))
          (if (fx<= Value (? Vector mid))
              (set! right mid)
              (set! left (fx+ mid 1)))))))

(define (ParMerge Source Left1 Right1 Left2 Right2 Target Left Depth)
  (if (fx= Depth 0)
      (Merge Source Left1 Right1 Left2 Right2 Target Left)
      (let ((size1 (fx+ (fx- Right1 Left1) 1))
            (size2 (fx+ (fx- Right2 Left2) 1)))
        (if (fx< size1 size2)
            (ParMerge Source Left2 Right2 Left1 Right1 Target Left Depth) 
            (cond
              ((fx> size1 0)
               (let* ((mid1 (fxquotient (fx+ Left1 Right1) 2))
                      (mid2 (BinarySearch (? Source mid1) Source Left2 Right2))
                      (mid (fx+ Left (fx- mid1 Left1) (fx- mid2 Left2))))
                 (! Target mid (? Source mid1))
                 (let ((f (future (λ () (ParMerge Source Left1 (fx- mid1 1)
                                                  Left2 (fx- mid2 1) Target Left (fx- Depth 1))))))
                   (ParMerge Source (fx+ mid1 1) Right1
                             mid2 Right2 Target (fx+ mid 1) (fx- Depth 1))
                   (touch f)))))))))

(define (ParSort Source Left Right Target Depth) 
  (cond
    ((fx= Depth 0)
     (Sort Source Left Right Target)) 
    ((fx< Left Right)
     (let* ((mid (fxquotient (fx+ Left Right) 2)))
       (define f (future (λ () (ParSort Target Left mid Source (fx- Depth 1)))))
       (ParSort Target (fx+ mid 1) Right Source (fx- Depth 1))
       (touch f)
       (ParMerge Source Left mid (fx+ mid 1) Right Target Left (fx- Depth 1)))))) 
   
(define (ParMergeSort Source MaxDepth)
  (let* ((size (fxvector-length Source))
         (target (make-fxvector size)))
    (do ((i 0 (fx+ i 1))) ((fx>= i size))
      (! target i (? Source i))) 
    (ParSort target 0 (fx- size 1) Source MaxDepth)))

(define size (* 16 1024 1024)) 
(define T (make-fxvector size))
(define ref 0)
(define MaxDepth 5)
(display "ParMergeSort of ")
(display size)
(display " integers") 
(newline)
(do ((depth 0 (fx+ depth 1))) ((fx> depth MaxDepth)) 
  (display "  spawning depth = ") 
  (display depth)
  (do ((i 0 (fx+ i 1))
       (x 1 (fxremainder (fx+ x 4253171) 1235711))) ((fx>= i size))
    (! T i x))
  (let ((t (get-time)))
    (ParMergeSort T depth)
    (set! t (- (get-time) t))
    (cond ((zero? depth) (set! ref t)))
    (display " - elapsed time = ")
    (display (~r t #:precision 2 #:min-width 5  #:pad-string " "))
    (display " - speed up = ")
    (display (~r (/ ref t)  #:precision 2 #:min-width 5  #:pad-string " "))
    (do ((i 1 (fx+ i 1))) ((fx>= i size))
      (cond ((fx> (? T (fx- i 1)) (? T i)) (error))))
    (newline)))
