#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <pthread.h>

static struct timeval start_time;

static float get_time(struct timeval start_time)
  { static struct timeval now_time;
    float time;
    gettimeofday(&now_time, NULL);
    time = now_time.tv_sec - start_time.tv_sec;
    time += (now_time.tv_usec - start_time.tv_usec)*1.0e-06;
    return time; }

//

static void q_sort(int Vector[], int Left, int Right)
  { int left = Left;
    int right = Right;
    int pivot = Vector[(left + right) / 2];
    while (left < right)
      { while (Vector[left]  < pivot) left  += 1;
        while (Vector[right] > pivot) right -= 1;
        if (left <= right)
          { int save = Vector[left];
            Vector[left++] = Vector[right];
            Vector[right--] = save; }}
    if (Left < right)
      q_sort(Vector, Left, right);
    if (Right > left)
      q_sort(Vector, left, Right); }

float QuickSort(int Source[], int Size)
  { float time;
    gettimeofday(&start_time, NULL);
    q_sort(Source, 0, Size - 1);
    time = get_time(start_time);
    return time;  }

//

void s_merge(int Source[], int Left1, int Right1,
             int Left2, int Right2, int Target[], int Left)
  { while (Left1 <= Right1 && Left2 <= Right2)
      if (Source[Left1] < Source[Left2])
        Target[Left++] = Source[Left1++];
      else
        Target[Left++] = Source[Left2++];
    while (Left1 <= Right1)
        Target[Left++] = Source[Left1++];
    while (Left2 <= Right2)
      Target[Left++] = Source[Left2++]; }

static void s_sort(int Source[], int Left, int Right, int Target [])
  { if (Left < Right)
      { int mid = (Left + Right) / 2;
        s_sort(Target, Left, mid, Source);
        s_sort(Target, mid + 1, Right, Source);
        s_merge(Source, Left, mid, mid + 1, Right, Target, Left); }}

float SeqMergeSort(int Source[], int Size)
  { int * target;
    float time;
    gettimeofday(&start_time, NULL);
    target = (int *)malloc(Size*sizeof(int));
    for (int i = 0; i < Size; i += 1)
      target[i] = Source[i];
    s_sort(target, 0, Size - 1, Source);
    free((void *)target);
    time = get_time(start_time);
    return time; }

//

static int binsearch(int Value, int Vector[], int Left, int Right)
  { if (Left > Right)
      return Left;
    Right += 1;
    while(Left < Right)
      { int mid = (Left + Right) / 2;
        if (Value <= Vector[mid])
          Right = mid;
        else
          Left  = mid + 1; }
    return Right; }

typedef struct { int * Source;
                 int   Left1;
                 int   mid1;
                 int   Left2;
                 int   mid2;
                 int * Target;
                 int   Left;
                 int   Depth; } MergeArgs;

static void p_merge(int [], int , int , int , int , int [], int , int );

static void merge_thread(MergeArgs * Arg)
  { p_merge(Arg->Source, Arg->Left1, Arg->mid1, Arg->Left2,
            Arg->mid2, Arg->Target, Arg->Left, Arg->Depth); }

static void p_merge(int Source[], int Left1, int Right1,
                    int Left2, int Right2, int Target[], int Left, int Depth)
  { if (Depth == 0)
      s_merge(Source, Left1, Right1, Left2, Right2, Target, Left);
    else
      { int size1 = Right1 - Left1 + 1;
        int size2 = Right2 - Left2 + 1;
        if (size1 < size2)
          p_merge(Source, Left2, Right2, Left1, Right1, Target, Left, Depth);
        else if (size1 > 0)
          { int mid1 = (Left1 + Right1) / 2;
            int mid2 = binsearch(Source[mid1], Source, Left2, Right2);
            int mid = Left + (mid1 - Left1) + (mid2 - Left2);
            pthread_t thread;
            MergeArgs arg;
            arg.Source = Source;
            arg.Left1 = Left1;
            arg.mid1 = mid1 - 1;
            arg.Left2 = Left2;
            arg.mid2 = mid2 - 1;
            arg.Target = Target;
            arg.Left = Left;
            arg.Depth = Depth - 1;
            Target[mid] = Source[mid1];
            assert(pthread_create(&thread, NULL, (void *)merge_thread, &arg) == 0);
            p_merge(Source, mid1 + 1, Right1, mid2 , Right2, Target, mid + 1, Depth - 1);
            assert(pthread_join(thread, NULL) == 0); }}}

typedef struct { int * Target;
                 int   Left;
                 int   mid;
                 int * Source;
                 int   Depth; } SortArgs;

static void p_sort(int[], int, int, int [], int);

static void SortThreadStart(SortArgs * Arg)
  { p_sort(Arg->Target, Arg->Left, Arg->mid, Arg->Source, Arg->Depth); }

static void p_sort(int Source[], int Left, int Right, int Target [], int Depth)
  { if (Depth == 0)
      s_sort(Source, Left, Right, Target);
    else if (Left < Right)
      { int mid = (Left + Right) / 2;
        pthread_t thread;
        SortArgs arg;
        arg.Target = Target;
        arg.Left = Left;
        arg.mid = mid;
        arg.Source = Source;
        arg.Depth = Depth - 1;
        assert(pthread_create(&thread, NULL, (void *)SortThreadStart, &arg) == 0);
        p_sort(Target, mid + 1, Right, Source, Depth - 1);
        assert(pthread_join(thread, NULL) == 0);
        p_merge(Source, Left, mid, mid + 1, Right, Target, Left, Depth-1); }}

float ParMergeSort(int Source[], int Size, int MaxDepth)
  { int * target;
    float time;
    gettimeofday(&start_time, NULL);
    target = (int *)malloc(Size*sizeof(int));
    for (int i = 0; i < Size; i += 1)
      target[i] = Source[i];
    p_sort(target, 0, Size - 1, Source, MaxDepth);
    free((void *)target);
    time = get_time(start_time);
    return time; }
